"""Module providing abstract Node2Vec implementation."""
from typing import Optional
from ensmallen import Graph
import numpy as np
from .ensmallen_embedder import EnsmallenEmbedder


class SPINE(EnsmallenEmbedder):
    """Abstract class for Node2Vec algorithms."""

    def __init__(
        self,
        embedding_size: int = 400,
        dtype: Optional[str] = "u8",
        verbose: bool = True
    ):
        """Create new abstract Node2Vec method.

        Parameters
        --------------------------
        embedding_size: int = 400
            Dimension of the embedding.
            Since the memory footprint of the SPINE algorithms
            is so low, we can easily affort to creater larger embeddings
            by default.
        dtype: Optional[str] = "u8"
            Dtype to use for the embedding. Note that an improper dtype may cause overflows.
        verbose: bool = True
            Whether to show loading bars
        """
        self._dtype = dtype

        super().__init__(
            embedding_size=embedding_size,
            verbose=verbose
        )

    def _fit_transform_graph(self, graph: Graph) -> np.ndarray:
        return graph.compute_spine_embedding(
            embedding_size=self._embedding_size,
            dtype=self._dtype,
            verbose=self._verbose,
        )



"""Abstract Keras Sequence object for running models on huge datasets."""
from keras_mixed_sequence import Sequence


class AbstractSequence(Sequence):
    """Abstract Keras Sequence object for running models on huge datasets."""

    def __init__(
        self,
        batch_size: int,
        sample_number: int,
        window_size: int = 4,
        elapsed_epochs: int = 0,
        random_state: int = 42
    ):
        """Create new Sequence object.

        Parameters
        -----------------------------
        batch_size: int,
            Number of nodes to include in a single batch.
        sample_number: int,
            Number of samples that compose this Sequence.
        window_size: int = 4,
            Window size for the local context.
            On the borders the window size is trimmed.
        elapsed_epochs: int = 0,
            Number of elapsed epochs to init state of generator.
        random_state: int = 42,
            Random random_state to make the sequence reproducible.
        """
        self._window_size = window_size
        self._random_state = random_state

        super().__init__(
            sample_number=sample_number,
            batch_size=batch_size,
            elapsed_epochs=elapsed_epochs
        )



"""Sub-module with utilities to make experiments easier to write."""
from .parameter_validators import validate_verbose, validate_window_size
from .gpu_utilities import has_gpus, has_nvidia_drivers, has_rocm_drivers, has_single_gpu, execute_gpu_checks, get_available_gpus_number
from .normalize_model_structural_parameters import normalize_model_list_parameter, normalize_model_ragged_list_parameter

try:
    from .tensorflow_utils import (
        tensorflow_version_is_higher_or_equal_than,
        must_have_tensorflow_version_higher_or_equal_than
    )
    from .graph_to_sparse_tensor import graph_to_sparse_tensor
    __all__ = [
        "graph_to_sparse_tensor",
        "validate_verbose",
        "validate_window_size",
        "has_gpus",
        "has_nvidia_drivers",
        "has_rocm_drivers",
        "has_single_gpu",
        "execute_gpu_checks",
        "get_available_gpus_number",
        "tensorflow_version_is_higher_or_equal_than",
        "must_have_tensorflow_version_higher_or_equal_than",
        "normalize_model_list_parameter", "normalize_model_ragged_list_parameter"
    ]
except ModuleNotFoundError:
    __all__ = [
        "validate_verbose",
        "validate_window_size",
        "has_gpus",
        "has_nvidia_drivers",
        "has_rocm_drivers",
        "has_single_gpu",
        "execute_gpu_checks",
        "get_available_gpus_number",
        "normalize_model_list_parameter", "normalize_model_ragged_list_parameter"
    ]



"""Submodule providing utils relative to embedders and graph embedding sanitization."""
from .enforce_sorted_graph import enforce_sorted_graph
from .detect_graph_node_embedding_oddities import detect_graph_node_embedding_oddities

__all__ = [
    "enforce_sorted_graph",
    "detect_graph_node_embedding_oddities"
]



"""NodeTransformer class to convert nodes to edge embeddings."""
from typing import List, Union
import numpy as np
import pandas as pd


class NodeTransformer:
    """NodeTransformer class to convert nodes to edge embeddings."""

    def __init__(
        self,
        numeric_node_ids: bool = False,
        aligned_node_mapping: bool = False,
    ):
        """Create new NodeTransformer object.

        Parameters
        -------------------
        numeric_node_ids: bool = False,
            Wether to return the numeric node IDs instead of the node embedding.
        aligned_node_mapping: bool = False,
            This parameter specifies whether the mapping of the embeddings nodes
            matches the internal node mapping of the given graph.
            If these two mappings do not match, the generated edge embedding
            will be meaningless.
        """
        self._numeric_node_ids = numeric_node_ids
        self._embedding = None
        self._embedding_numpy = None
        self._aligned_node_mapping = aligned_node_mapping

    @property
    def numeric_node_ids(self) -> bool:
        """Return whether the transformer returns numeric node IDs."""
        return self._numeric_node_ids

    def fit(self, embedding: pd.DataFrame):
        """Fit the model.

        Parameters
        -------------------------
        embedding: pd.DataFrame,
            Embedding to use to fit the transformer.
            This is a pandas DataFrame and NOT a numpy array because we need
            to be able to remap correctly the vector embeddings in case of
            graphs that do not respect the same internal node mapping but have
            the same node set. It is possible to remap such graphs using
            Ensmallen's remap method but it may be less intuitive to users.
        """
        if not isinstance(embedding, pd.DataFrame):
            if not self._aligned_node_mapping:
                raise ValueError("Given embedding is not a pandas DataFrame.")
            self._embedding_numpy = embedding
        else:
            self._embedding = embedding
            self._embedding_numpy = embedding.to_numpy()

    def transform(self, nodes: Union[List[str], List[int]]) -> np.ndarray:
        """Return embeddings from given node.

        Parameters
        --------------------------
        nodes: Union[List[str], List[int]],
            List of nodes whose embedding is to be returned.
            By default this should be a list of strings, if the
            aligned_node_mapping is setted, then this methods also accepts
            a list of ints.

        Raises
        --------------------------
        ValueError,
            If embedding is not fitted.

        Returns
        --------------------------
        Numpy array of embeddings.
        """
        if self._embedding_numpy is None and not self.numeric_node_ids:
            raise ValueError(
                "Transformer was not fitted yet."
            )

        if self._aligned_node_mapping:
            if self.numeric_node_ids:
                return nodes
            return self._embedding_numpy[nodes]

        if self.numeric_node_ids:
            return self._embedding.index.get_indexer(nodes)
        return self._embedding.loc[nodes].to_numpy()



"""Submodule providing pipelines for edge prediction."""
from typing import Union, Callable, Tuple, List, Optional, Dict
import pandas as pd
import tensorflow as tf
from ensmallen import Graph
from tqdm.auto import trange, tqdm
import math
from ensmallen.datasets import get_dataset
from yaml import warnings
from ..edge_prediction import Perceptron, MultiLayerPerceptron, EdgePredictionModel
from ..utils import execute_gpu_checks, get_available_gpus_number
from .compute_node_embedding import compute_node_embedding

edge_prediction_models = {
    "Perceptron": Perceptron,
    "MultiLayerPerceptron": MultiLayerPerceptron
}


def get_negative_graphs(
    graph: Graph,
    random_state: int,
    sample_only_edges_with_heterogeneous_node_types: bool,
    unbalance_rate: float,
    train_size: float,
) -> Tuple[Graph, Graph]:
    """Return tuple with training and test negative graphs.

    Implementative details
    ----------------------------
    We generate first one larger graph of negative edges,
    and then we split it into a training and test subgraphs
    using the `random hodout` schema, that is we do not preserve
    the number of connected components as we do not care about them
    since we are not using the topology. Do note that this may
    not be the same case when using a GCN for edge prediction.

    Parameters
    ----------------------------
    graph: Graph
        The graph from which to sample the negative edges.
    random_state: int
        Random state to sample the negatives and execute the random split.
    sample_only_edges_with_heterogeneous_node_types: bool
        Whether to only sample edges between heterogeneous node types.
        This may be useful when training a model to predict between
        two portions in a bipartite graph.
    unbalance_rate: float
        Quantity over over (or under) sampling for the negatives respectively
        to the number of edges in the provided graph.
    train_size: float
        Split size of the training graph.
    """
    # For both the training and the test set we sample the same
    # number of negative edges are there are existing edges in the training and test graphs, respectively.
    # This is done in order to avoid an excessive bias in the evaluation of the edge prediction.
    # Across multiple holdouts, statistically it is unlikely to sample
    # consistently unknown positive edges, and therefore we should be able
    # to remove this negative bias from the evaluation.
    # Of course, this only apply to graphs where we can assume that there is
    # not a massive amount of unknown positive edges.
    graph.enable()
    negative_graph = graph.sample_negatives(
        number_of_negative_samples=int(math.ceil(graph.get_edges_number()*unbalance_rate)),
        random_state=random_state,
        sample_only_edges_with_heterogeneous_node_types=sample_only_edges_with_heterogeneous_node_types,
        verbose=False
    )
    negative_graph.enable()
    return negative_graph.random_holdout(
        train_size=train_size,
        random_state=random_state,
        verbose=False,
    )


def evaluate_embedding_for_edge_prediction(
    embedding_method: Union[str, Callable[[Graph, int], pd.DataFrame]],
    graphs: Union[Graph, str, List[str], List[Graph]],
    model_name: Union[str, Callable[[Graph, pd.DataFrame], EdgePredictionModel]],
    epochs: int = 1000,
    number_of_holdouts: int = 10,
    train_size: float = 0.8,
    unbalance_rates: Tuple[Union[float, str]] = (10.0, 100.0, "auto"),
    random_state: int = 42,
    batch_size: int = 2**10,
    edge_embedding_method: str = "Concatenate",
    trainable_embedding: bool = False,
    use_dropout: bool = True,
    dropout_rate: float = 0.1,
    use_edge_metrics: bool = False,
    edge_types: Optional[List[str]] = None,
    use_mirrored_strategy: Union[bool, str] = "auto",
    use_only_cpu: Union[bool, str] = "auto",
    only_execute_embeddings: bool = False,
    embedding_method_fit_kwargs: Optional[Dict] = None,
    embedding_method_kwargs: Optional[Dict] = None,
    subgraph_of_interest_for_edge_prediction: Optional[Graph] = None,
    sample_only_edges_with_heterogeneous_node_types: bool = False,
    devices: Union[List[str], str] = None,
    graph_normalization_callback: Callable[[Graph], Graph] = None,
    verbose: bool = True
) -> Tuple[pd.DataFrame, List[pd.DataFrame]]:
    """Return the evaluation of an embedding for edge prediction on the given model.

    Parameters
    ----------------------
    embedding_method: Union[str, Callable[[Graph, int], pd.DataFrame]]
        Either the name of the embedding method or a method to be called.
    graphs: Union[Graph, str, List[str], List[Graph]]
        The graph to run the embedding and edge prediction on.
        If a string was provided, we will retrieve the graphs from Ensmallen's repositories.
        If a list was provided, we will iterate on all graphs.
    model_name: Union[str, Callable[[Graph, pd.DataFrame], EdgePredictionModel]]
        Either the name of the model or a method returning a model.
    epochs: int = 1000
        Number of epochs to train the perceptron model for.
    number_of_holdouts: int = 10
        The number of the holdouts to run.
    train_size: float = 0.8
        Split size of the training graph.
    unbalance_rates: Tuple[Union[float, str]] = (10.0, 100.0, "auto")
        List of unbalance to evaluate.
        Do note that an unbalance of one is always included.
        With "auto", we use the true unbalance of the graph.
    random_state: int = 42
        The seed to be used to reproduce the holdout.
    batch_size: int = 2**10
        Size of the batch to be considered.
    edge_types: Optional[List[str]] = None
        Edge types to focus the edge prediction on, if any.
    use_mirrored_strategy: Union[bool, str] = "auto"
        Whether to use mirror strategy to distribute the
        computation across multiple devices.
        This is automatically enabled if more than one
        GPU is detected and the flag `use_only_gpu` was
        not provided, or if the list of devices to use
        was provided and it includes at least a GPU.
    use_only_cpu: Union[bool, str] = "auto",
        Whether to only use CPU.
        Do note that for CBOW and SkipGram models,
        this will switch the implementation from the
        TensorFlow implementation and will use our Rust Ensmallen one.
    only_execute_embeddings: bool = False
        Whether to only execute the computation of the embedding or also the edge prediction.
        This flag can be useful when the two operations should be executed on different machines
        or at different times.
    embedding_method_fit_kwargs: Optional[Dict] = None
        The kwargs to be forwarded to the embedding fit method
    embedding_method_kwargs: Optional[Dict] = None
        The kwargs to be forwarded to the embedding method
    subgraph_of_interest_for_edge_prediction: Optional[Graph] = None
        The subgraph to use for the edge prediction training and evaluation, if any.
        If none are provided, we sample the negative edges from the entire graph.
    sample_only_edges_with_heterogeneous_node_types: bool = False
        Whether to only sample edges between heterogeneous node types.
        This may be useful when training a model to predict between
        two portions in a bipartite graph.
    devices: Union[List[str], str] = None
        The list of devices to use when training the embedding and edge prediction models
        in a MirroredStrategy, that is across multiple GPUs. Thise feature is mainly useful
        when there are multiple GPUs available AND the graph is large enough to actually
        use the GPUs (for instance when it has at least a few million nodes).
    graph_normalization_callback: Callable[[Graph], Graph] = None
        Graph normalization procedure to call on graphs that have been loaded from
        the Ensmallen automatic retrieval.
    verbose: bool = True
        Whether to show the loading bars.
    """
    if isinstance(model_name, str) and model_name not in edge_prediction_models:
        raise ValueError(
            f"The given edge prediction model `{model_name}` is not supported. "
            f"The supported node embedding methods are `{edge_prediction_models}`."
        )

    if embedding_method_kwargs is None:
        embedding_method_kwargs = {}

    if subgraph_of_interest_for_edge_prediction is not None and not subgraph_of_interest_for_edge_prediction.has_edges():
        raise ValueError(
            "The provided subgraph of interest does not have any edge."
        )

    # If devices are given as a single device we adapt this into a list.
    if isinstance(devices, str):
        devices = [devices]

    # If in the list of provided devices there is a GPU specified,
    # and there are more than one GPU, we need to use the MirroredStrategy
    # to distribute its computation.
    if devices and any(
        "GPU" in device
        for device in devices
    ) and get_available_gpus_number() > 1:
        use_mirrored_strategy = True

    # If the embedding method is a string, we execute this check also within
    # the compute node embedding pipeline.
    if not isinstance(embedding_method, str):
        execute_gpu_checks(use_mirrored_strategy)

    if isinstance(graphs, (Graph, str)):
        graphs = [graphs]


    holdouts = []
    histories = []
    for graph in tqdm(
        graphs,
        desc="Executing graph",
        disable=len(graphs) <= 1,
        dynamic_ncols=True,
        leave=False
    ):
        if isinstance(graph, str):
            graph = get_dataset(graph)()
            if graph_normalization_callback is not None:
                graph = graph_normalization_callback(graph)
        
        graph_name = graph.get_name()
        graph_unbalance_rate = graph.get_nodes_number() * (graph.get_nodes_number() - 1) / graph.get_number_of_directed_edges()

        unbalance_rates_for_graph = [
            graph_unbalance_rate
            if unbalance_rate == "auto"
            else unbalance_rate
            for unbalance_rate in unbalance_rates
            if unbalance_rate != "auto" and unbalance_rate < graph_unbalance_rate
        ]

        if len(unbalance_rates_for_graph) < len(unbalance_rates):
            warnings.warn(
                (
                    "Be advised that the provided unbalance rates included "
                    "rates that were higher than the maximum possible unbalance rate "
                    "possible in this graph {:4f}."
                ).format(
                    graph_unbalance_rate
                )
            )

        for holdout_number in trange(
            number_of_holdouts,
            desc=f"Executing holdouts on {graph_name}",
            dynamic_ncols=True,
            leave=False
        ):
            train_graph, test_graph = graph.connected_holdout(
                train_size,
                random_state=random_state*holdout_number,
                edge_types=edge_types,
                verbose=False
            )

            if isinstance(embedding_method, str):
                embedding, _ = compute_node_embedding(
                    graph=train_graph,
                    node_embedding_method_name=embedding_method,
                    use_mirrored_strategy=use_mirrored_strategy,
                    use_only_cpu=use_only_cpu,
                    fit_kwargs=embedding_method_fit_kwargs,
                    devices=devices,
                    verbose=False,
                    **embedding_method_kwargs
                )
            else:
                embedding = embedding_method(
                    train_graph,
                    holdout_number,
                    use_mirrored_strategy=use_mirrored_strategy,
                    **embedding_method_kwargs
                )

            if only_execute_embeddings:
                continue

            # If requested, we focus the training and test graphs
            # into the area of interest.
            if subgraph_of_interest_for_edge_prediction is not None:
                # We remove all the nodes from the training and test graphs
                # which are not part of the subgraph of interest.
                subgraph_node_names = subgraph_of_interest_for_edge_prediction.get_node_names()
                train_graph = train_graph.filter_from_names(
                    node_names_to_keep=subgraph_node_names
                )
                test_graph = test_graph.filter_from_names(
                    node_names_to_keep=subgraph_node_names
                )

                # We remove all the edges from the training and test graphs
                # which are not part of the subgraph of interest.
                train_graph = train_graph & subgraph_of_interest_for_edge_prediction
                test_graph = test_graph & subgraph_of_interest_for_edge_prediction

                assert train_graph.has_edges()
                assert test_graph.has_edges()

                # We realign the considered training and test graph with
                # the provided subgraph of interest node names dictionary.
                train_graph = train_graph.remap_from_graph(
                    subgraph_of_interest_for_edge_prediction
                )
                test_graph = test_graph.remap_from_graph(
                    subgraph_of_interest_for_edge_prediction
                )

                assert train_graph.has_edges()
                assert test_graph.has_edges()

            graph_to_use_to_sample_negatives = (
                # We sample the negative edges from the entire graph
                graph
                # when no subgraph of interest has been provided
                if subgraph_of_interest_for_edge_prediction is None
                # else, if the subgraph was provided, we only extract the negative
                # edges from this portion of the graph, making the evaluation of the edge prediction task
                # more significant for the actual desired task: often, when doing an edge predition
                # in a graph, we actually intend to predict some type of edges of interest.
                # When this is the case, we do not care to train or evaluate the model on edges that
                # are not in the portion of the graph.
                # Consider a graph representing a social network: if we are interested in learning the
                # connections between users living in small towns, if we also take into account users
                # living in large metropolis we would both train the model on unrelevant data and evaluate
                # the model of a different task, which may be more or less difficult.
                else subgraph_of_interest_for_edge_prediction
            )

            # Force alignment
            embedding = embedding.loc[train_graph.get_node_names()]

            # Simplify the train graph if needed.
            if train_graph.has_edge_weights():
                train_graph.remove_inplace_edge_weights()

            if train_graph.has_edge_types():
                train_graph.remove_inplace_edge_types()

            # Simplify the test graph if needed.
            if test_graph.has_edge_weights():
                test_graph.remove_inplace_edge_weights()

            if test_graph.has_edge_types():
                test_graph.remove_inplace_edge_types()

            train_graph.enable(
                vector_sources=True,
                vector_destinations=True,
                vector_cumulative_node_degrees=True
            )

            train_negative_graph, test_negative_graph = get_negative_graphs(
                graph_to_use_to_sample_negatives,
                # Inside ensmallen we pass this number in a splitmix function
                # therefore even a sum should have quite enough additional entropy.
                random_state=random_state + holdout_number,
                sample_only_edges_with_heterogeneous_node_types=sample_only_edges_with_heterogeneous_node_types,
                unbalance_rate=1.0,
                train_size=train_size,
            )

            if tf.config.list_physical_devices('GPU') and use_mirrored_strategy:
                strategy = tf.distribute.MirroredStrategy(devices)
            else:
                # Use the Default Strategy
                strategy = tf.distribute.get_strategy()

            with strategy.scope():
                if isinstance(model_name, str):
                    model = edge_prediction_models.get(model_name)(
                        graph=graph,
                        embedding=embedding,
                        edge_embedding_method=edge_embedding_method,
                        trainable_embedding=trainable_embedding,
                        use_dropout=use_dropout,
                        dropout_rate=dropout_rate,
                        use_edge_metrics=use_edge_metrics,
                    )
                else:
                    model = model_name(graph, embedding)

            histories.append(model.fit(
                train_graph=train_graph,
                valid_graph=test_graph,
                negative_valid_graph=test_negative_graph,
                batch_size=batch_size,
                epochs=epochs,
                sample_only_edges_with_heterogeneous_node_types=sample_only_edges_with_heterogeneous_node_types,
                verbose=False
            ))

            train_performance = model.evaluate(
                graph=train_graph,
                negative_graph=train_negative_graph,
                batch_size=batch_size,
                verbose=False
            )
            train_performance["evaluation_type"] = "train"
            train_performance["unbalance"] = 1.0
            train_performance["graph_name"] = graph_name

            test_performance = model.evaluate(
                graph=test_graph,
                negative_graph=test_negative_graph,
                batch_size=batch_size,
                verbose=False
            )
            test_performance["evaluation_type"] = "test"
            test_performance["unbalance"] = 1.0
            test_performance["graph_name"] = graph_name

            if isinstance(embedding_method, str):
                train_performance["embedding_method"] = embedding_method
                test_performance["embedding_method"] = embedding_method

            holdouts.append(train_performance)
            holdouts.append(test_performance)

            for i, unbalance_rate in enumerate(tqdm(
                unbalance_rates_for_graph,
                desc="Evaluating on datasets with different unbalance",
                dynamic_ncols=True,
                leave=False
            ), start=1):
                train_negative_graph, test_negative_graph = get_negative_graphs(
                    graph_to_use_to_sample_negatives,
                    # Inside ensmallen we pass this number in a splitmix function
                    # therefore even a sum should have quite enough additional entropy.
                    random_state=random_state + holdout_number + i,
                    sample_only_edges_with_heterogeneous_node_types=sample_only_edges_with_heterogeneous_node_types,
                    unbalance_rate=unbalance_rate,
                    train_size=train_size,
                )
                train_performance = model.evaluate(
                    graph=train_graph,
                    negative_graph=train_negative_graph,
                    batch_size=batch_size,
                    verbose=False
                )
                train_performance["evaluation_type"] = "train"
                train_performance["unbalance"] = unbalance_rate
                train_performance["graph_name"] = graph_name
                
                test_performance = model.evaluate(
                    graph=test_graph,
                    negative_graph=test_negative_graph,
                    batch_size=batch_size,
                    verbose=False
                )
                test_performance["evaluation_type"] = "test"
                test_performance["unbalance"] = unbalance_rate
                test_performance["graph_name"] = graph_name

                holdouts.append(train_performance)
                holdouts.append(test_performance)

                if isinstance(embedding_method, str):
                    train_performance["embedding_method"] = embedding_method
                    test_performance["embedding_method"] = embedding_method


    return pd.DataFrame(holdouts), histories



"""GraphCBOW model for graph embedding."""
from typing import Dict, Union, Optional

import numpy as np
import pandas as pd
from ensmallen import Graph
from tensorflow.keras.optimizers import \
    Optimizer   # pylint: disable=import-error,no-name-in-module

from .cbow import CBOW
from .node2vec import Node2Vec

class GraphCBOW(Node2Vec):
    """GraphCBOW model for graph embedding.

    The GraphCBOW model for graoh embedding receives a list of contexts and tries
    to predict the central word. The model makes use of an NCE loss layer
    during the training process to generate the negatives.
    """

    def __init__(
        self,
        graph: Graph,
        embedding_size: int = 100,
        embedding: Union[np.ndarray, pd.DataFrame] = None,
        optimizer: Union[str, Optimizer] = None,
        number_of_negative_samples: int = 5,
        walk_length: int = 128,
        batch_size: int = 256,
        iterations: int = 16,
        window_size: int = 4,
        return_weight: float = 1.0,
        explore_weight: float = 1.0,
        change_node_type_weight: float = 1.0,
        change_edge_type_weight: float = 1.0,
        max_neighbours: Optional[int] = 100,
        elapsed_epochs: int = 0,
        random_state: int = 42,
        dense_node_mapping: Optional[Dict[int, int]] = None,
        use_gradient_centralization: bool = True,
        siamese: bool = False
    ):
        """Create new sequence TensorFlowEmbedder model.

        Parameters
        -------------------------------------------
        graph: Graph,
            Graph to be embedded.
        word2vec_model: Word2Vec,
            Word2Vec model to use.
        embedding_size: int = 100,
            Dimension of the embedding.
        embedding: Union[np.ndarray, pd.DataFrame] = None,
            The seed embedding to be used.
            Note that it is not possible to provide at once both
            the embedding and either the vocabulary size or the embedding size.
        optimizer: Union[str, Optimizer] = None,
            The optimizer to be used during the training of the model.
            By default, if None is provided, Nadam with learning rate
            set at 0.01 is used.
        window_size: int = 4,
            Window size for the local context.
            On the borders the window size is trimmed.
        number_of_negative_samples: int = 5,
            The number of negative classes to randomly sample per batch.
            This single sample of negative classes is evaluated for each element in the batch.
        walk_length: int = 128,
            Maximal length of the walks.
        batch_size: int = 256,
            Number of nodes to include in a single batch.
        iterations: int = 16,
            Number of iterations of the single walks.
        window_size: int = 4,
            Window size for the local context.
            On the borders the window size is trimmed.
        return_weight: float = 1.0,
            Weight on the probability of returning to the same node the walk just came from
            Having this higher tends the walks to be
            more like a Breadth-First Search.
            Having this very high  (> 2) makes search very local.
            Equal to the inverse of p in the Node2Vec paper.
        explore_weight: float = 1.0,
            Weight on the probability of visiting a neighbor node
            to the one we're coming from in the random walk
            Having this higher tends the walks to be
            more like a Depth-First Search.
            Having this very high makes search more outward.
            Having this very low makes search very local.
            Equal to the inverse of q in the Node2Vec paper.
        change_node_type_weight: float = 1.0,
            Weight on the probability of visiting a neighbor node of a
            different type than the previous node. This only applies to
            colored graphs, otherwise it has no impact.
        change_edge_type_weight: float = 1.0,
            Weight on the probability of visiting a neighbor edge of a
            different type than the previous edge. This only applies to
            multigraphs, otherwise it has no impact.
        max_neighbours: Optional[int] = 100,
            Number of maximum neighbours to consider when using approximated walks.
            By default, None, we execute exact random walks.
            This is mainly useful for graphs containing nodes with extremely high degrees.
        elapsed_epochs: int = 0,
            Number of elapsed epochs to init state of generator.
        random_state: int = 42,
            The random state to reproduce the training sequence.
        dense_node_mapping: Optional[Dict[int, int]] = None,
            Mapping to use for converting sparse walk space into a dense space.
            This object can be created using the method (available from the
            graph object created using Graph)
            called `get_dense_node_mapping` that returns a mapping from
            the non trap nodes (those from where a walk could start) and
            maps these nodes into a dense range of values.
        use_gradient_centralization: bool = True,
            Whether to wrap the provided optimizer into a normalized
            one that centralizes the gradient.
            It is automatically enabled if the current version of
            TensorFlow supports gradient transformers.
            More detail here: https://arxiv.org/pdf/2004.01461.pdf
        siamese: bool = False
            Whether to use the siamese modality and share the embedding
            weights between the source and destination nodes.
        """
        super().__init__(
            graph=graph,
            word2vec_model=CBOW,
            embedding_size=embedding_size,
            embedding=embedding,
            optimizer=optimizer,
            number_of_negative_samples=number_of_negative_samples,
            walk_length=walk_length,
            batch_size=batch_size,
            iterations=iterations,
            window_size=window_size,
            return_weight=return_weight,
            explore_weight=explore_weight,
            change_node_type_weight=change_node_type_weight,
            change_edge_type_weight=change_edge_type_weight,
            max_neighbours=max_neighbours,
            elapsed_epochs=elapsed_epochs,
            random_state=random_state,
            dense_node_mapping=dense_node_mapping,
            use_gradient_centralization=use_gradient_centralization,
            siamese=siamese
        )



"""Model implementing Node Label Neighbour for graphs."""
from typing import Dict, Tuple, Union, Optional

import numpy as np
import pandas as pd
from ensmallen import Graph
from extra_keras_metrics import get_minimal_multiclass_metrics
from keras_mixed_sequence import MixedSequence
from sklearn.preprocessing import RobustScaler
from tensorflow.keras import regularizers  # pylint: disable=import-error,no-name-in-module
from tensorflow.keras.layers import (  # pylint: disable=import-error,no-name-in-module
    Dense,  # pylint: disable=import-error,no-name-in-module
    Dropout,  # pylint: disable=import-error,no-name-in-module
    Embedding,  # pylint: disable=import-error,no-name-in-module
    Concatenate,  # pylint: disable=import-error,no-name-in-module
    GlobalAveragePooling1D,  # pylint: disable=import-error,no-name-in-module
    BatchNormalization,  # pylint: disable=import-error,no-name-in-module
    Input  # pylint: disable=import-error,no-name-in-module
)
from tensorflow.keras.models import Model  # pylint: disable=import-error,no-name-in-module
from tensorflow.keras.losses import CategoricalCrossentropy  # pylint: disable=import-error,no-name-in-module
from tensorflow.keras.optimizers import Optimizer  # pylint: disable=import-error,no-name-in-module

from ..sequences import NodeLabelPredictionSequence
from ..embedders.tensorflow_embedders import TensorFlowEmbedder


class NodeLabelPredictionfeedForwardNeuralNetwork(TensorFlowEmbedder):
    """Class implementing Feed-Forward Neural Network for node-label prediction."""

    def __init__(
        self,
        graph: Graph,
        labels_number: int = None,
        node_embedding_size: int = None,
        use_node_embedding_dropout: bool = True,
        node_embedding_dropout_rate: float = 0.3,
        use_node_features_dropout: bool = True,
        node_features_dropout_rate: float = 0.6,
        use_batch_normalization: bool = True,
        l1_kernel_regularization: float = 1e-3,
        l2_kernel_regularization: float = 1e-3,
        hidden_dense_layers: Union[int, Tuple[int]] = (),
        node_embedding: Union[np.ndarray, pd.DataFrame] = None,
        node_features: Union[np.ndarray, pd.DataFrame] = None,
        optimizer: Union[str, Optimizer] = "nadam",
        trainable_node_embedding: bool = False,
        scaler: "Scaler" = "RobustScaler"
    ):
        """Create new NoLaN model.

        Parameters
        -------------------
        TODO!
        """
        self._graph = graph
        self._labels_number = self._graph.get_node_types_number(
        ) if labels_number is None else labels_number
        self._use_node_embedding_dropout = use_node_embedding_dropout
        self._node_embedding_dropout_rate = node_embedding_dropout_rate
        self._use_node_features_dropout = use_node_features_dropout
        self._node_features_dropout_rate = node_features_dropout_rate
        self._use_batch_normalization = use_batch_normalization
        self._l1_kernel_regularization = l1_kernel_regularization
        self._l2_kernel_regularization = l2_kernel_regularization
        if isinstance(hidden_dense_layers, int):
            hidden_dense_layers = (hidden_dense_layers, )
        self._hidden_dense_layers = hidden_dense_layers

        if scaler == "RobustScaler":
            scaler = RobustScaler()

        if node_embedding is not None:
            scaled_node_embedding = scaler.fit_transform(node_embedding)
            if isinstance(node_embedding, pd.DataFrame):
                node_embedding = pd.DataFrame(
                    scaled_node_embedding,
                    columns=node_embedding.columns,
                    index=node_embedding.index,
                )

        self._node_features_size = None

        if node_features is not None:
            if node_embedding is not None:
                if isinstance(node_features, pd.DataFrame) != isinstance(node_embedding, pd.DataFrame):
                    raise ValueError(
                        "Node embedding and node features must either be both "
                        "pandas DataFrames or both numpy arrays."
                    )
                if node_features.shape[0] != node_embedding.shape[0]:
                    raise ValueError(
                        (
                            "Given node features must be available"
                            " for each of the {} nodes.".format(
                                self._vocabulary_size)
                        )
                    )
                if isinstance(node_features, pd.DataFrame) and (node_features.index != node_embedding.index).any():
                    raise ValueError(
                        "Index of node features and node embedding must match!"
                    )
            if scaler is not None:
                scaled_node_features = scaler.fit_transform(node_features)
                if isinstance(node_features, pd.DataFrame):
                    node_features = pd.DataFrame(
                        scaled_node_features,
                        columns=node_features.columns,
                        index=node_features.index,
                    )
            self._node_features_size = node_features.shape[1]
        self._node_features = node_features

        super().__init__(
            vocabulary_size=graph.get_nodes_number() if node_embedding is None else None,
            embedding_size=node_embedding_size if node_embedding is None else None,
            embedding=node_embedding,
            optimizer=optimizer,
            trainable_embedding=trainable_node_embedding
        )

    def _build_model(self):
        """Return NoLaN model."""
        node_star_input = Input((None,), name="NodeStarInput")

        node_embedding_layer = Embedding(
            input_dim=self._vocabulary_size+1,
            output_dim=self._embedding_size,
            weights=None if self._embedding is None else [np.vstack([
                np.zeros(self._embedding_size),
                self._embedding
            ])],
            mask_zero=True,
            name=TensorFlowEmbedder.TERMS_EMBEDDING_LAYER_NAME
        )

        mean_node_star_embedding = GlobalAveragePooling1D(
            name="MeanNodeStarEmbedding"
        )(
            node_embedding_layer(node_star_input),
            mask=node_embedding_layer.compute_mask(node_star_input),
        )

        if self._use_node_embedding_dropout:
            mean_node_star_embedding = Dropout(
                self._node_embedding_dropout_rate,
                name="NodeEmbeddingDropout"
            )(mean_node_star_embedding)

        if self._node_features is not None:
            node_features_layer = Embedding(
                input_dim=self._vocabulary_size+1,
                output_dim=self._node_features_size,
                weights=None if self._embedding is None else [np.vstack([
                    np.zeros(self._node_features_size),
                    self._node_features
                ])],
                mask_zero=True,
                trainable=False,
                name="NodeFeatures"
            )

            mean_node_star_features = GlobalAveragePooling1D(
                name="MeanNodeStarFeatures"
            )(
                node_features_layer(node_star_input),
                mask=node_features_layer.compute_mask(node_star_input),
            )

            if self._use_node_features_dropout:
                mean_node_star_features = Dropout(
                    self._node_features_dropout_rate,
                    name="NodeFeaturesDropout"
                )(mean_node_star_features)

            mean_node_star_embedding = Concatenate(
                name="NodedataConcatenation"
            )((
                mean_node_star_embedding,
                mean_node_star_features
            ))

        if self._use_batch_normalization:
            mean_node_star_embedding = BatchNormalization(
                scale=False,
                center=False
            )(mean_node_star_embedding)

        hidden = mean_node_star_embedding

        for units in self._hidden_dense_layers:
            hidden = Dense(
                units,
                kernel_regularizer=regularizers.l1_l2(
                    l1=self._l1_kernel_regularization,
                    l2=self._l2_kernel_regularization
                ),
                activation="relu"
            )(hidden)

        output = Dense(
            self._labels_number,
            kernel_regularizer=regularizers.l1_l2(
                l1=self._l1_kernel_regularization,
                l2=self._l2_kernel_regularization
            ),
            activation="softmax"
        )(hidden)

        model = Model(
            inputs=node_star_input,
            outputs=output,
            name="NoLaN"
        )

        return model

    def _compile_model(self) -> Model:
        """Compile model."""
        self._model.compile(
            optimizer=self._optimizer,
            loss=CategoricalCrossentropy(label_smoothing=0.1),
            metrics=get_minimal_multiclass_metrics()
        )

    def build_training_sequence(
        self,
        train_graph: Graph,
        max_neighbours: Optional[int] = 100,
        batch_size: int = 256,
        validation_graph: Graph = None,
        random_state: int = 42
    ) -> Tuple[MixedSequence, MixedSequence]:
        """Return .

        Parameters
        --------------------
        train_graph: Graph
            Training graph.
        max_neighbours: Optional[int] = 100,
            Number of neighbours to consider.
            If None, the graph median is used.
        batch_size: int = 256,
            Batch size for the sequence.
        validation_graph: Graph = None,
            The graph whose nodes are to be predicted.
        random_state: int = 42,
            Random seed to reproduce.

        Returns
        --------------------
        Training and validation MixedSequence
        """
        train_sequence = NodeLabelPredictionSequence(
            train_graph,
            max_neighbours=max_neighbours,
            batch_size=batch_size,
            random_state=random_state,
        )
        if validation_graph is not None:
            validation_sequence = NodeLabelPredictionSequence(
                validation_graph,
                max_neighbours=validation_graph.get_max_degree(),
                batch_size=batch_size,
                random_state=random_state,
            )
        else:
            validation_sequence = None
        return train_sequence, validation_sequence

    def fit(
        self,
        train_graph: Graph,
        max_neighbours: Optional[int] = 100,
        batch_size: int = 256,
        epochs: int = 1000,
        validation_graph: Graph = None,
        early_stopping_monitor: str = "loss",
        early_stopping_min_delta: float = 0,
        early_stopping_patience: int = 30,
        early_stopping_mode: str = "min",
        reduce_lr_monitor: str = "loss",
        reduce_lr_min_delta: float = 0,
        reduce_lr_patience: int = 5,
        reduce_lr_mode: str = "min",
        reduce_lr_factor: float = 0.9,
        verbose: int = 1,
        random_state: int = 42,
        **kwargs: Dict
    ) -> pd.DataFrame:
        """Return pandas dataframe with training history.

        Parameters
        -----------------------
        train_graph: Graph,
            Training graphs.
        max_neighbours: Optional[int] = 100,
            Number of neighbours to consider.
            If None, the graph median is used.
        epochs: int = 10000,
            Epochs to train the model for.
        validation_graph: Graph = None,
            Data reserved for validation.
        early_stopping_monitor: str = "loss",
            Metric to monitor for early stopping.
        early_stopping_min_delta: float = 0.1,
            Minimum delta of metric to stop the training.
        early_stopping_patience: int = 5,
            Number of epochs to wait for when the given minimum delta is not
            achieved after which trigger early stopping.
        early_stopping_mode: str = "min",
            Direction of the variation of the monitored metric for early stopping.
        reduce_lr_monitor: str = "loss",
            Metric to monitor for reducing learning rate.
        reduce_lr_min_delta: float = 1,
            Minimum delta of metric to reduce learning rate.
        reduce_lr_patience: int = 3,
            Number of epochs to wait for when the given minimum delta is not
            achieved after which reducing learning rate.
        reduce_lr_mode: str = "min",
            Direction of the variation of the monitored metric for learning rate.
        reduce_lr_factor: float = 0.9,
            Factor for reduction of learning rate.
        verbose: int = 1,
            Wethever to show the loading bar.
            Specifically, the options are:
            * 0 or False: No loading bar.
            * 1 or True: Showing only the loading bar for the epochs.
            * 2: Showing loading bar for both epochs and batches.
        **kwargs: Dict,
            Additional kwargs to pass to the Keras fit call.

        Returns
        -----------------------
        Dataframe with training history.
        """
        train_sequence, validation_sequence = self.build_training_sequence(
            train_graph,
            max_neighbours=max_neighbours,
            batch_size=batch_size,
            validation_graph=validation_graph,
            random_state=random_state
        )
        return super().fit(
            train_sequence,
            epochs=epochs,
            validation_graph=validation_sequence,
            early_stopping_monitor=early_stopping_monitor,
            early_stopping_min_delta=early_stopping_min_delta,
            early_stopping_patience=early_stopping_patience,
            early_stopping_mode=early_stopping_mode,
            reduce_lr_monitor=reduce_lr_monitor,
            reduce_lr_min_delta=reduce_lr_min_delta,
            reduce_lr_patience=reduce_lr_patience,
            reduce_lr_mode=reduce_lr_mode,
            reduce_lr_factor=reduce_lr_factor,
            verbose=verbose,
            **kwargs
        )

    def predict(
        self,
        X: Graph,
        verbose: bool = False,
        batch_size=256,
        random_state: int = 42
    ):
        """Run predict."""
        return self._model.predict(
            NodeLabelPredictionSequence(
                X,
                max_neighbours=self._graph.max_degree(),
                batch_size=batch_size,
                random_state=random_state,
            ),
            verbose=verbose
        )

    def evaluate(
        self,
        graph: Graph,
        validation_graph: Graph = None,
        verbose: bool = False,
        batch_size=256,
        random_state: int = 42
    ) -> pd.DataFrame:
        """Run predict.

        TODO! Update docstring!
        """
        train_sequence, validation_sequence = self.build_training_sequence(
            graph,
            max_neighbours=self._graph.max_degree(),
            batch_size=batch_size,
            validation_graph=validation_graph,
            random_state=random_state
        )
        return pd.DataFrame(
            [{
                **dict(zip(
                    self._model.metrics_names,
                    self._model.evaluate(
                        train_sequence,
                        verbose=verbose
                    )
                )),
                "run_type": "training"
            },
                *(({
                    **dict(zip(
                        self._model.metrics_names,
                        self._model.evaluate(
                            validation_sequence,
                            verbose=verbose
                        )
                    )),
                    "run_type": "validation"
                }, ) if validation_sequence is not None
                else ()
            )]
        )

    @property
    def embedding(self) -> np.ndarray:
        """Return model embeddings.

        Raises
        -------------------
        NotImplementedError,
            If the current embedding model does not have an embedding layer.
        """
        # We need to drop the first column (feature) of the embedding
        # curresponding to the indices 0, as this value is reserved for the
        # masked values. The masked values are the values used to fill
        # the batches of the neigbours of the nodes.
        return TensorFlowEmbedder.embedding.fget(self)[1:]  # pylint: disable=no-member

    def get_embedding_dataframe(self) -> pd.DataFrame:
        """Return terms embedding using given index names."""
        return super().get_embedding_dataframe(self._graph.get_node_names())



"""Keras Sequence for running Neural Network on graph edge prediction."""
from typing import Union, Tuple, List, Optional
import numpy as np

from ensmallen import Graph

from embiggen.sequences.edge_label_prediction_sequence import EdgeLabelPredictionSequence  # pylint: disable=no-name-in-module


class BinaryEdgeLabelPredictionSequence(EdgeLabelPredictionSequence):
    """Keras Sequence for running a Neural Network on graph edge-label prediction."""

    def __init__(
        self,
        graph: Graph,
        positive_edge_type: Optional[Union[str, int]] = None,
        use_node_types: bool = False,
        use_edge_metrics: bool = False,
        batch_size: int = 2**10,
        graph_to_avoid: Graph = None,
        batches_per_epoch: Union[int, str] = "auto",
        elapsed_epochs: int = 0,
        random_state: int = 42
    ):
        """Create new EdgeLabekPredictionSequence object.

        Parameters
        --------------------------------
        graph: Graph
            The graph from which to sample the edges.
        positive_edge_type: Optional[Union[str, int]] = None
            The type for the positive class.
            If None, we check if the provided graph has two classes,
            and if so we use the minority class as the positive class.
            If the graph does not have two classes, we will raise an error.
        use_node_types: bool = False
            Whether to return the node types.
        use_edge_metrics: bool = False
            Whether to return the edge metrics.
        batch_size: int = 2**10
            The batch size to use.
        graph_to_avoid: Graph = None
            Graph to avoid when generating the edges.
            This can be the validation component of the graph, for example.
            More information to how to generate the holdouts is available
            in the Graph package.
        batches_per_epoch: Union[int, str] = "auto"
            Number of batches per epoch.
            If auto, it is used: `10 * edges number /  batch size`
        elapsed_epochs: int = 0
            Number of elapsed epochs to init state of generator.
        random_state: int = 42
            The random_state to use to make extraction reproducible.
        """
        if positive_edge_type is None:
            if graph.get_edge_types_number() != 2:
                raise ValueError(
                    "The provided edge types number is None, so we would use "
                    "as positive class the class with least edges, but the "
                    "graph you have provided does not have two classes and "
                    "therefore we cannot infer exactly which of the minority "
                    "classes you would want to use."
                )
            else:
                edge_types_counts = graph.get_edge_type_names_counts_hashmap()
                positive_edge_type = min(
                    edge_types_counts,
                    key=edge_types_counts.get
                )

        if not isinstance(positive_edge_type, (int, str)):
            raise ValueError(
                (
                    "The positive edge type must be either an integer "
                    "or a string but the provided one is a {}."
                ).format(
                    type(positive_edge_type)
                )
            )
        if positive_edge_type < 0:
            raise ValueError(
                (
                    "The positive edge type must be a positive integer "
                    "but the provided edge type was {}."
                ).format(
                    type(positive_edge_type)
                )
            )

        # If the provided positive edge type is a string, we need
        # to convert this into the corresponding numeric type.
        # Do note that within Ensmallen the type is checked for existance,
        # that is when a non existant type is provided a ValueError exception
        # will be raised.
        # Furthermore, if the current graph does not contain edge type at all
        # an exception will be raised.
        if isinstance(positive_edge_type, str):
            positive_edge_type = graph.get_edge_type_id_from_edge_type_name(
                positive_edge_type
            )

        # We check if the provided positive edge type
        if positive_edge_type >= graph.get_edge_types_number():
            raise ValueError(
                (
                    "The current graph has {} edge types but "
                    "the provided edge type was {}."
                ).format(
                    graph.get_edge_types_number(),
                    positive_edge_type
                )
            )

        self._positive_edge_type = positive_edge_type
        super().__init__(
            graph,
            use_node_types=use_node_types,
            use_edge_metrics=use_edge_metrics,
            batch_size=batch_size,
            graph_to_avoid=graph_to_avoid,
            batches_per_epoch=batches_per_epoch,
            elapsed_epochs=elapsed_epochs,
            random_state=random_state
        )

    def __getitem__(self, idx: int) -> Tuple[List[np.ndarray], np.ndarray]:
        """Return batch corresponding to given index.

        Parameters
        ---------------
        idx: int,
            Index corresponding to batch to be returned.

        Returns
        ---------------
        Return Tuple containing X and Y numpy arrays corresponding to given batch index.
        """
        edge_data, edge_types = super().__getitem__(idx)
        return edge_data, edge_types == self._positive_edge_type



"""Module providing abstract Node2Vec implementation."""
from typing import Optional
from ensmallen import Graph
import numpy as np
from ensmallen import models
from .ensmallen_embedder import EnsmallenEmbedder


class Node2Vec(EnsmallenEmbedder):
    """Abstract class for Node2Vec algorithms."""

    MODELS = {
        "cbow": models.CBOW,
        "skipgram": models.SkipGram
    }

    def __init__(
        self,
        model_name: str,
        embedding_size: int = 100,
        epochs: int = 10,
        clipping_value: float = 6.0,
        number_of_negative_samples: int = 5,
        log_sigmoid: Optional[bool] = True,
        siamese: Optional[bool] = False,
        walk_length: int = 128,
        iterations: int = 1,
        window_size: int = 4,
        return_weight: float = 1.0,
        explore_weight: float = 1.0,
        change_node_type_weight: float = 1.0,
        change_edge_type_weight: float = 1.0,
        max_neighbours: Optional[int] = 100,
        learning_rate: float = 0.01,
        learning_rate_decay: float = 0.9,
        normalize_by_degree: bool = False,
        random_state: int = 42,
        verbose: bool = True
    ):
        """Create new abstract Node2Vec method.

        Parameters
        --------------------------
        model_name: str
            The ensmallen graph embedding model to use.
            Can either be the SkipGram of CBOW models.
        embedding_size: int = 100
            Dimension of the embedding.
        epochs: int = 10
            Number of epochs to train the model for.
        window_size: int = 4
            Window size for the local context.
            On the borders the window size is trimmed.
        clipping_value: float = 6.0
            Value at which we clip the dot product, mostly for numerical stability issues.
            By default, `6.0`, where the loss is already close to zero.
        number_of_negative_samples: int = 5
            The number of negative classes to randomly sample per batch.
            This single sample of negative classes is evaluated for each element in the batch.
        walk_length: int = 128
            Maximal length of the walks.
        iterations: int = 1
            Number of iterations of the single walks.
        window_size: int = 4
            Window size for the local context.
            On the borders the window size is trimmed.
        return_weight: float = 1.0
            Weight on the probability of returning to the same node the walk just came from
            Having this higher tends the walks to be
            more like a Breadth-First Search.
            Having this very high  (> 2) makes search very local.
            Equal to the inverse of p in the Node2Vec paper.
        explore_weight: float = 1.0
            Weight on the probability of visiting a neighbor node
            to the one we're coming from in the random walk
            Having this higher tends the walks to be
            more like a Depth-First Search.
            Having this very high makes search more outward.
            Having this very low makes search very local.
            Equal to the inverse of q in the Node2Vec paper.
        change_node_type_weight: float = 1.0
            Weight on the probability of visiting a neighbor node of a
            different type than the previous node. This only applies to
            colored graphs, otherwise it has no impact.
        change_edge_type_weight: float = 1.0
            Weight on the probability of visiting a neighbor edge of a
            different type than the previous edge. This only applies to
            multigraphs, otherwise it has no impact.
        max_neighbours: Optional[int] = 100
            Number of maximum neighbours to consider when using approximated walks.
            By default, None, we execute exact random walks.
            This is mainly useful for graphs containing nodes with high degrees.
        learning_rate: float = 0.01
            The learning rate to use to train the Node2Vec model.
        learning_rate_decay: float = 0.9
            Factor to reduce the learning rate for at each epoch. By default 0.9.
        normalize_by_degree: bool = False
            Whether to normalize the random walk by the node degree
            of the destination node degrees.
        random_state: int = 42
            The random state to reproduce the training sequence.
        verbose: bool = True
            Whether to show loading bars
        """
        if model_name.lower() not in ("cbow", "skipgram"):
            raise ValueError(
                (
                    "The provided model name {} is not supported. "
                    "The supported models are `CBOW` and `SkipGram`."
                ).format(model_name)
            )
        self._epochs = epochs
        self._learning_rate = learning_rate
        self._learning_rate_decay = learning_rate_decay

        self._model = Node2Vec.MODELS[model_name.lower()](
            embedding_size=embedding_size,
            window_size=window_size,
            clipping_value=clipping_value,
            number_of_negative_samples=number_of_negative_samples,
            log_sigmoid=log_sigmoid,
            siamese=siamese,
            walk_length=walk_length,
            return_weight=return_weight,
            explore_weight=explore_weight,
            change_edge_type_weight=change_edge_type_weight,
            change_node_type_weight=change_node_type_weight,
            max_neighbours=max_neighbours,
            random_state=random_state,
            iterations=iterations,
            normalize_by_degree=normalize_by_degree
        )

        super().__init__(
            embedding_size=embedding_size,
            verbose=verbose
        )

    def _fit_transform_graph(self, graph: Graph) -> np.ndarray:
        return self._model.fit_transform(
            graph, 
            epochs=self._epochs,
            learning_rate=self._learning_rate,
            learning_rate_decay=self._learning_rate_decay,
            verbose=self._verbose
        )


"""Sub-module with methods to compute node-embedding with a one-liner."""
import inspect
from typing import Dict, List, Tuple, Union

import pandas as pd
from cache_decorator import Cache
from ensmallen import Graph

from ..utils import execute_gpu_checks, get_available_gpus_number, has_gpus

from ..embedders import SUPPORTED_NODE_EMBEDDING_METHODS
from ..embedders.ensmallen_embedders import EnsmallenEmbedder

REQUIRE_ZIPFIAN = [
    "cbow",
    "skipgram"
]

RANDOM_WALK_BASED_MODELS = [
    "cbow",
    "glove",
    "skipgram",
    "spine"
]

LINK_PREDICTION_BASED_MODELS = [
    "siamese",
    "transr",
    "transe",
    "transh",
    "simple"
]


def get_available_node_embedding_methods() -> List[str]:
    """Return list of supported node embedding methods."""
    return list(SUPPORTED_NODE_EMBEDDING_METHODS.keys())


def get_node_embedding_method(
    node_embedding_method_name: str,
    use_only_cpu: bool
):
    """Return node embedding method curresponding to given name.

    Parameters
    ---------------------
    node_embedding_method_name: str
        The name of the embedding method to retrieve
    use_only_cpu: bool
        Whether to retrieve CPU or GPU versions of the model, when available.
    """
    model = SUPPORTED_NODE_EMBEDDING_METHODS[node_embedding_method_name]
    # If there is a further choice to be made for this model.
    if isinstance(model, dict):
        model = model["cpu"] if use_only_cpu else model["gpu"]
    return model


def is_node_embedding_method_supported(node_embedding_method_name: str) -> bool:
    """Return boolean value representing if given node embedding method is supported.

    Parameters
    --------------------
    node_embedding_method_name: str,
        Name of the node embedding method.

    Returns
    --------------------
    Whether the given node embedding method is supported.
    """
    return node_embedding_method_name in get_available_node_embedding_methods()


@Cache(
    cache_path=[
        "node_embeddings/{node_embedding_method_name}/tensorflow/{graph_name}/{_hash}_embedding.pkl.gz",
        "node_embeddings/{node_embedding_method_name}/tensorflow/{graph_name}/{_hash}_training_history.csv.xz",
    ],
    args_to_ignore=["devices", "use_mirrored_strategy", "verbose"]
)
def _compute_node_tensorflow_embedding(
    graph: Graph,
    graph_name: str,  # pylint: disable=unused-argument
    node_embedding_method_name: str,
    fit_kwargs: Dict,
    verbose: bool,
    use_mirrored_strategy: bool,
    devices: Union[List[str], str],
    **kwargs: Dict
) -> Tuple[Union[pd.DataFrame, Tuple[pd.DataFrame]], pd.DataFrame]:
    """Wrapper for `compute_node_embedding` method."""
    # Since the verbose kwarg may be provided also on the fit_kwargs
    # we normalize the parameter to avoid collisions.
    verbose = fit_kwargs.pop("verbose", verbose)

    node_embedding_model = get_node_embedding_method(
        node_embedding_method_name,
        False
    )
    # Otherwise it is a TensorFlow-based model.
    import tensorflow as tf
    if tf.config.list_physical_devices('GPU') and use_mirrored_strategy:
        strategy = tf.distribute.experimental.MultiWorkerMirroredStrategy(
            devices
        )
    else:
        strategy = tf.distribute.get_strategy()

    with strategy.scope():
        # Creating the node embedding model
        model = node_embedding_model(
            graph=graph,
            **kwargs
        )

    # Fitting the node embedding model
    history = model.fit(
        verbose=verbose,
        **fit_kwargs
    )
    return model.get_embedding_dataframe(), history


@Cache(
    cache_path="node_embeddings/{node_embedding_method_name}/ensmallen/{graph_name}/{_hash}_embedding.csv.gz",
    args_to_ignore=["verbose"]
)
def _compute_node_ensmallen_embedding(
    graph: Graph,
    graph_name: str,  # pylint: disable=unused-argument
    node_embedding_method_name: str,
    fit_kwargs: Dict,
    verbose: bool,
    **kwargs: Dict
) -> Tuple[Union[pd.DataFrame, Tuple[pd.DataFrame]], pd.DataFrame]:
    """Wrapper for `compute_node_embedding` method."""
    # Since the verbose kwarg may be provided also on the fit_kwargs
    # we normalize the parameter to avoid collisions.
    verbose = fit_kwargs.pop("verbose", verbose)

    node_embedding_model = get_node_embedding_method(
        node_embedding_method_name,
        True
    )

    # Create the Ensmallen-based model.
    model = node_embedding_model(
        verbose=verbose,
        **kwargs
    )
    return model.fit_transform_graph(
        graph,
        **fit_kwargs
    )


def compute_node_embedding(
    graph: Graph,
    node_embedding_method_name: str,
    use_only_cpu: Union[bool, str] = "auto",
    use_mirrored_strategy: Union[bool, str] = "auto",
    devices: Union[List[str], str] = None,
    fit_kwargs: Dict = None,
    verbose: Union[bool, int] = True,
    automatically_drop_unsupported_parameters: bool = False,
    automatically_enable_time_memory_tradeoffs: bool = True,
    **kwargs: Dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Return embedding computed using SkipGram on given graph.

    Parameters
    --------------------------
    graph: Graph,
        Graph to embed.
    node_embedding_method_name: str,
        The name of the node embedding method to use.
    use_only_cpu: Union[bool, str] = "auto",
        Whether to only use CPU.
        Do note that for CBOW and SkipGram models,
        this will switch the implementation from the
        TensorFlow implementation and will use our Rust Ensmallen one.
    use_mirrored_strategy: Union[bool, str] = "auto"
        Whether to use mirror strategy to distribute the
        computation across multiple devices.
        This is automatically enabled if more than one
        GPU is detected and the flag `use_only_gpu` was
        not provided, or if the list of devices to use
        was provided and it includes at least a GPU.
    devices: Union[List[str], str] = None,
        The devices to use.
        If None, all GPU devices available are used.
    fit_kwargs: Dict = None,
        Arguments to pass to the fit call.
    verbose: bool = True,
        Whether to show loading bars.
    automatically_drop_unsupported_parameters: bool = False,
        If required, we filter out the unsupported parameters.
        This may be useful when running a suite of experiments with a set of
        parameters and you do not want to bother in dropping the parameters
        that are only supported in a subset of methods.
    automatically_enable_time_memory_tradeoffs: bool = True,
        Whether to activate the time memory tradeoffs automatically.
        Often case, this is something you want enabled on your graph object.
        Since, generally, it is a good idea to enable these while
        computing a node embedding we enable these by default.
    **kwargs: Dict,
        Arguments to pass to the node embedding method constructor.
        Read the documentation of the selected method to learn
        which methods are supported by the selected constructor.

    Returns
    --------------------------
    Tuple with node embedding and training history.
    """
    lower_node_embedding_method_name = node_embedding_method_name.lower()
    if not is_node_embedding_method_supported(lower_node_embedding_method_name):
        raise ValueError(
            (
                "The given node embedding method `{}` is not supported. "
                "The supported node embedding methods are `{}`."
            ).format(
                node_embedding_method_name,
                get_available_node_embedding_methods()
            )
        )

    # If devices are given as a single device we adapt this into a list.
    if isinstance(devices, str):
        devices = [devices]

    # If in the list of provided devices there is a GPU specified,
    # and there are more than one GPU, we need to use the MirroredStrategy
    # to distribute its computation.
    if devices and any(
        "GPU" in device
        for device in devices
    ) and get_available_gpus_number() > 1:
        use_mirrored_strategy = True

    # If the use only CPU parameter is set to auto,
    # we automatically enable it when no GPUs are
    # available.
    if use_only_cpu == "auto" and not has_gpus():
        use_only_cpu = True

    if not use_only_cpu and use_mirrored_strategy == "auto" and get_available_gpus_number() > 1:
        use_mirrored_strategy = True

    if use_only_cpu and use_mirrored_strategy == True:
        raise ValueError(
            "It does not make sense to require to use only CPU "
            "and require to use the mirrored strategy for GPUs."
        )

    execute_gpu_checks(use_mirrored_strategy)

    # If the fit kwargs are not given we normalize them to an empty dictionary.
    if fit_kwargs is None:
        fit_kwargs = {}

    # If required, we filter out the unsupported parameters.
    # This may be useful when running a suite of experiments with a set of
    # parameters and you do not want to bother in dropping the parameters
    # that are only supported in a subset of methods.
    if automatically_drop_unsupported_parameters:
        if kwargs:
            # Get the list of supported parameters
            supported_parameter = inspect.signature(
                get_node_embedding_method(
                    lower_node_embedding_method_name,
                    use_only_cpu
                ).__init__
            ).parameters
            # Filter out the unsupported parameters
            kwargs = {
                key: value
                for key, value in kwargs.items()
                if key in supported_parameter
            }
        if fit_kwargs:
            # Get the list of supported parameters
            model = get_node_embedding_method(
                lower_node_embedding_method_name,
                use_only_cpu
            )
            if issubclass(model, EnsmallenEmbedder):
                supported_parameter = inspect.signature(
                    model.fit_transform_graph
                ).parameters
            else:
                supported_parameter = inspect.signature(
                    model.fit
                ).parameters
            # Filter out the unsupported parameters
            fit_kwargs = {
                key: value
                for key, value in kwargs.items()
                if key in supported_parameter
            }

    # If required we enable the time memory tradeoffs.
    if automatically_enable_time_memory_tradeoffs:
        if lower_node_embedding_method_name in RANDOM_WALK_BASED_MODELS:
            graph.enable(
                vector_sources=False,
                vector_destinations=True,
                vector_cumulative_node_degrees=True
            )
        if lower_node_embedding_method_name in LINK_PREDICTION_BASED_MODELS:
            graph.enable(
                vector_sources=True,
                vector_destinations=True,
                vector_cumulative_node_degrees=False
            )

    if issubclass(get_node_embedding_method(
        lower_node_embedding_method_name,
        use_only_cpu
    ), EnsmallenEmbedder):
        # Call the wrapper with cache.
        # Do note that this model does not return any history.
        return _compute_node_ensmallen_embedding(
            graph,
            graph_name=graph.get_name(),
            node_embedding_method_name=lower_node_embedding_method_name,
            fit_kwargs=fit_kwargs,
            verbose=verbose,
            **kwargs
        ), None
    return _compute_node_tensorflow_embedding(
        graph,
        graph_name=graph.get_name(),
        node_embedding_method_name=lower_node_embedding_method_name,
        fit_kwargs=fit_kwargs,
        verbose=verbose,
        use_mirrored_strategy=use_mirrored_strategy,
        devices=devices,
        **kwargs
    )



"""Module with models for graph and text embedding and their Keras Sequences."""
import warnings
try:
    from .embedders.tensorflow_embedders import (
        CBOW, GloVe, GraphCBOW, GraphGloVe, GraphSkipGram,
        SkipGram, TransE, TransH, TransR, SimplE, Siamese
    )
    from .node_prediction import GraphConvolutionalNeuralNetwork
    from .sequences import Word2VecSequence
    from .transformers import CorpusTransformer

    all_exports = [
        "CBOW", "SkipGram", "GloVe", "GraphGloVe", "Word2VecSequence", "CorpusTransformer",
        "TransE", "TransH", "TransR", "SimplE", "Siamese", "GraphConvolutionalNeuralNetwork"
    ]
except ModuleNotFoundError as e:
    from .embedders.ensmallen_embedders import *
    all_exports = []
    warnings.warn(
        "You do not have TensorFlow installed, all tensorflow models "
        "will not work. We have replaced some default models, namely GraphCBOW "
        "and GraphSkipGram with the Ensmallen Rust versions."
    )

from .embedders.ensmallen_embedders import SPINE
from .transformers import (EdgeTransformer,
                           GraphTransformer, LinkPredictionTransformer,
                           NodeTransformer)
from .visualizations import GraphVisualization

__all__ = [
    "GraphCBOW",
    "GraphSkipGram",
    "NodeTransformer",
    "EdgeTransformer",
    "GraphTransformer",
    "LinkPredictionTransformer",
    "GraphVisualization",
    "SPINE",
    *all_exports
]



"""Submodule providing methods to enforce the graph sorting."""
from ensmallen import Graph


def enforce_sorted_graph(graph: Graph):
    """Checks whether the graph has nodes sorted properly.

    Parameters
    ----------------
    graph: Graph
        The graph to run checks on.

    Raises
    ----------------
    ValueError
        If the graph does not have graph sorted by decreasing
        node degree.
    """
    if not graph.has_nodes_sorted_by_decreasing_outbound_node_degree():
        raise ValueError(
            "The given graph does not have the nodes sorted by decreasing "
            "order, therefore the negative sampling (which follows a zipfian "
            "distribution) would not approximate well the Softmax.\n"
            "In order to sort the given graph in such a way that the node IDs "
            "are sorted by decreasing outbound node degrees, you can use "
            "the Graph method "
            "`graph.sort_by_decreasing_outbound_node_degree()`"
        )



"""Submodule providing methods to check for graph oddities."""
import warnings
from ensmallen import Graph


def detect_graph_node_embedding_oddities(graph: Graph):
    """Raises warnings if there are topological oddities in this graph.

    Parameters
    ----------------
    graph: Graph
        The graph to run checks on.

    Raises
    ----------------
    ValueError
        If the provided graph does not have nodes.
    ValueError
        If the provided graph does not have edges.
    """
    if not graph.has_nodes():
        raise ValueError("The provided graph does not have nodes.")

    if not graph.has_edges():
        raise ValueError("The provided graph does not have edges.")
    
    if graph.has_singleton_nodes():
        warnings.warn(
            (
                "Please be advised that this graph contains {} singleton nodes. "
                "Consider that node embedding algorithms that only use topological information "
                "such as CBOW, GloVe, SPINE and SkipGram are not able to provide meaningful "
                "embeddings for these nodes, and their embedding will be generally "
                "far away from any other node. It is also possible that all singleton nodes "
                "will receive a relatively similar node embedding. "
                "Consider dropping them by using the `graph.drop_singleton_nodes()` method."
            ).format(graph.get_singleton_nodes_number())
        )
    
    if graph.has_singleton_nodes_with_selfloops():
        warnings.warn(
            (
                "Please be advised that this graph contains {} singleton nodes with self-loops. "
                "Consider that node embedding algorithms that only use topological information "
                "such as CBOW, GloVe, SPINE and SkipGram are not able to provide meaningful "
                "embeddings for these nodes, and their embedding will be generally "
                "far away from any other node. It is also possible that all singleton nodes "
                "will receive a relatively similar node embedding. "
                "Consider dropping them by using the `graph.drop_singleton_nodes_with_selfloops()` method."
            ).format(graph.get_singleton_nodes_with_selfloops_number())
        )


"""Submodule providing experiments pipelines."""
import warnings
from .compute_node_embedding import (
    compute_node_embedding,
    get_available_node_embedding_methods
)
try:
    # TODO!: add non-tensorflow models to the edge embedding pipeline
    # such as the models from Sklearn.
    from .edge_prediction import evaluate_embedding_for_edge_prediction
    __all__ = [
        "compute_node_embedding",
        "get_available_node_embedding_methods",
        "evaluate_embedding_for_edge_prediction"
    ]
except:
    __all__ = [
        "compute_node_embedding",
        "get_available_node_embedding_methods",
    ]



"""Submodule providing abstract class for embedding graph nodes using Ensmallen."""
from typing import Union, Dict, Optional
from ensmallen import Graph
from ..embedders_utils import detect_graph_node_embedding_oddities
from ...utils import validate_verbose
import pandas as pd
import numpy as np


class EnsmallenEmbedder:
    """Abstract class providing graph node embedders based on Ensmallen."""

    def __init__(
        self,
        embedding_size: int = 100,
        verbose: bool = True
    ):
        """Create new EnsmallenEmbedder object.
        
        Parameters
        --------------------
        embedding_size: int = 100
            Number of dimensions of the node embedding
        verbose: bool = True
            Whether to show loading bars

        Raises
        --------------------
        ValueError
            If the provided embedding size is less than 1.
        """
        if not isinstance(embedding_size, int) or embedding_size < 1:
            raise ValueError(
                (
                    "The embedding size must be a strictly positive integer, "
                    "but you have provided {}."
                ).format(embedding_size)
            )
        validate_verbose(verbose)
        
        self._embedding_size = embedding_size
        self._verbose = verbose

    def _fit_transform_graph(
        self,
        graph: Graph,
    ) -> np.ndarray:
        """Method to be reimplemented in child classes to implement.
        
        Parameters
        -------------------
        graph: Graph
            The ensmallen graph to embed.
        """
        raise NotImplementedError(
            "The method `_fit_transform_graph` must be implemented "
            "in child classes."
        )

    def fit_transform_graph(
        self,
        graph: Graph,
        return_dataframe: bool = True
    ) -> Union[pd.DataFrame, np.ndarray]:
        """Return node embedding for provided graph.

        Parameters
        ---------------------
        graph: Graph
            The ensmallen graph to embed.
        return_dataframe: bool = True
            Whether to return the dataframe or directly
            a numpy array. Do note that using a DataFrame
            increases non trivially the amount of memory
            required for a graph node embedding.
        """
        detect_graph_node_embedding_oddities(graph)
        node_embedding = self._fit_transform_graph(graph)
        if return_dataframe:
            return pd.DataFrame(
                node_embedding,
                index=graph.get_node_names()
            )
        return node_embedding


"""Class for creating a Feed-Forward Neural Network model for edge prediction tasks."""
from typing import Union, List, Optional
from ensmallen import Graph
from tensorflow.keras import regularizers  # pylint: disable=import-error,no-name-in-module
from tensorflow.keras.layers import Dense, Layer  # pylint: disable=import-error,no-name-in-module

import numpy as np
import pandas as pd

from .edge_prediction_model import EdgePredictionModel


class FeedForwardNeuralNetwork(EdgePredictionModel):

    def __init__(
        self,
        graph: Graph,
        units: Optional[List[int]] = None,
        embedding_size: int = None,
        embedding: Union[np.ndarray, pd.DataFrame] = None,
        edge_embedding_method: str = "Concatenate",
        optimizer: str = "nadam",
        trainable_embedding: bool = False,
        use_dropout: bool = True,
        dropout_rate: float = 0.5,
        use_edge_metrics: bool = False,
        task_name: str = "EDGE_PREDICTION"
    ):
        """Create new Feed Forward Neural Network object.

        Parameters
        --------------------
        graph: Graph,
            The graph object to base the model on.
        units: Optional[List[int]] = None
            Units to use to build the hidden layers of the model.
        embedding_size: int = None
            Dimension of the embedding.
            If None, the seed embedding must be provided.
            It is not possible to provide both at once.
        embedding: Union[np.ndarray, pd.DataFrame] = None
            The seed embedding to be used.
            Note that it is not possible to provide at once both
            the embedding and either the vocabulary size or the embedding size.
        edge_embedding_method: str = "Concatenate",
            Method to use to create the edge embedding.
        optimizer: str = "nadam",
            Optimizer to use during the training.
        trainable_embedding: bool = False,
            Whether to allow for trainable embedding.
        use_dropout: bool = True,
            Whether to use dropout.
        dropout_rate: float = 0.5,
            Dropout rate.
        use_edge_metrics: bool = False,
            Whether to return the edge metrics.
        task_name: str = "EDGE_PREDICTION",
            The name of the task to build the model for.
            The currently supported task names are `EDGE_PREDICTION` and `EDGE_LABEL_PREDICTION`.
            The default task name is `EDGE_PREDICTION`.
        """
        self._units = [] if units is None else units
        super().__init__(
            graph,
            embedding_size=embedding_size,
            embedding=embedding,
            edge_embedding_method=edge_embedding_method,
            optimizer=optimizer,
            trainable_embedding=trainable_embedding,
            use_dropout=use_dropout,
            dropout_rate=dropout_rate,
            use_edge_metrics=use_edge_metrics,
            task_name=task_name
        )

    def _build_model_body(self, input_layer: Layer) -> Layer:
        """Build new model body for edge prediction."""
        for unit in self._units:
            input_layer = Dense(
                units=unit,
                activation="relu",
                kernel_regularizer=regularizers.l1_l2(l1=1e-5, l2=1e-4),
            )(input_layer)
        if self._task_name == "EDGE_PREDICTION":
            return Dense(
                units=1,
                activation="sigmoid",
                kernel_regularizer=regularizers.l1_l2(l1=1e-5, l2=1e-4),
            )(input_layer)
        elif self._task_name == "EDGE_LABEL_PREDICTION":
            return Dense(
                units=self._graph.get_edge_types_number(),
                activation="softmax",
                kernel_regularizer=regularizers.l1_l2(l1=1e-5, l2=1e-4),
            )(input_layer)
        else:
            ValueError("Unreacheable.")